/*
 * All modifications are copyright (c) 2000, K A Fraser.
 * 
 * Please note the original copyright and license details below.
 */

/*
 * Copyright (C) 1997 Massachusetts Institute of Technology 
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose and without fee or royalty is
 * hereby granted, provided that the full text of this NOTICE appears on
 * ALL copies of the software and documentation or portions thereof,
 * including modifications, that you make.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS," AND COPYRIGHT HOLDERS MAKE NO
 * REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED. BY WAY OF EXAMPLE,
 * BUT NOT LIMITATION, COPYRIGHT HOLDERS MAKE NO REPRESENTATIONS OR
 * WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR
 * THAT THE USE OF THE SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY
 * THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. COPYRIGHT
 * HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE OR
 * DOCUMENTATION.
 *
 * The name and trademarks of copyright holders may NOT be used in
 * advertising or publicity pertaining to the software without specific,
 * written prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with copyright
 * holders. See the file AUTHORS which should have accompanied this software
 * for a list of all copyright holders.
 *
 * This file may be derived from previously copyrighted software. This
 * copyright applies only to those changes made by the copyright
 * holders listed in the AUTHORS file. The rest of this file is covered by
 * the copyright notices, if any, listed below.
 */

#ifndef __DPF_H__
#define __DPF_H__

enum { DPF_BOGUSID=-1, DPF_OVERLAP=-2, DPF_NOMEM=-3 };

#define FILTER_SIZE 256*32  /* approx. bytes */

struct dpf_ir;

/* call this at start of day. */
void dpf_init(void);

/* initialize a filter structure. */
void dpf_begin(struct dpf_ir *ir);

/*
 * Compare message value to constant:
 *      msg[byte_offset:nbits] & mask == val
 */
void dpf_mkeq(struct dpf_ir *ir, uint8 nbits, uint16 offset, 
              uint32 mask, uint32 val);

/* 
 * Shift the base message pointer: 
 *	msg += (msg[byte_offset:nbits] & mask) << shift; 
 */
/*
 * KAF: new parameter `ext'. 
 *  == SHIFT:   shift message pointer
 *  == HDR_EXT: extend header length count by shift amount
 *  == PKT_EXT: extend both header and packet length counts by shift amount
 */
enum { SHIFT=1<<0, HDR_EXT=1<<1, PKT_EXT=1<<2 };
void dpf_mkshift(struct dpf_ir *ir, uint8 nbits, uint16 offset, 
                 uint32 mask, uint8 shift, uint8 ext);

/* Shift the base message pointer by a constant: msg += nbytes. */
void dpf_shifti(struct dpf_ir *ir, uint16 nbytes, uint8 ext);

/* Filter insertion/deletion. */
int dpf_insert(struct dpf_ir *filter, unsigned pid);
int dpf_delete(unsigned pid);

extern int (*dpf_iptr)(uint8 *msg, unsigned nbytes);

#endif /* __DPF_H__ */
